"""
The COOM parser.

Traverses the abstract syntax tree of the COOM input
in a visitor style fashion and outputs ASP facts.
"""

# flake8: noqa
# pylint: skip-file
# mypy: ignore-errors
import sys
from typing import List, Optional

try:
    from .coom_grammar.model.ModelParser import ModelParser
    from .coom_grammar.model.ModelVisitor import ModelVisitor
    from .coom_grammar.user.UserInputParser import UserInputParser
    from .coom_grammar.user.UserInputVisitor import UserInputVisitor
except ModuleNotFoundError:  # nocoverage
    print("COOM grammar files not found. Please run \n\n     ./build_grammar.sh")
    sys.exit(1)


class ASPUserInputVisitor(UserInputVisitor):
    """
    Custom visitor of the COOM Parser.
    Custom visitor of the COOM User Input Parser.
    Generates a list of ASP facts as strings.
    """

    def __init__(self) -> None:
        super().__init__()
        self.context: str = ""
        self.output_asp: List[str] = []

    def visitInput_block(self, ctx: UserInputParser.Input_blockContext):
        self.context = ctx.path().getText() + "."
        super().visitInput_block(ctx)
        self.context = ""

    def visitSet_value(self, ctx: UserInputParser.Set_valueContext):
        path = self.context + ctx.path().getText()
        value = ctx.formula_atom().getText()
        is_str = ctx.formula_atom().path() is not None
        if is_str:
            self.output_asp.append(f'user_value("root.{path}","{value}").')
        else:
            self.output_asp.append(f'user_value("root.{path}",{value}).')
        super().visitSet_value(ctx)

    def visitAdd_instance(self, ctx: UserInputParser.Add_instanceContext):
        path = self.context + ctx.path().getText()
        self.output_asp.append(f'user_include("root.{path}").')
        super().visitAdd_instance(ctx)


class ASPModelVisitor(ModelVisitor):
    """
    Custom visitor of the COOM Parser.
    Generates a list of ASP facts as strings.
    """

    def __init__(self) -> None:
        super().__init__()
        self.parent_enum: Optional[ModelParser.EnumerationContext] = None
        self.root_name: str = "product"
        self.structure_name: str = self.root_name
        self.context: str = self.root_name
        self.constraint_idx: int = 0
        self.condition_idx: int = 0
        self.row_idx: int = 0
        self.print_path: bool = True
        self.output_asp: List[str] = []

    def visitProduct(self, ctx: ModelParser.ProductContext):
        self.output_asp.append(f'structure("{self.root_name}").')
        super().visitProduct(ctx)

    def visitStructure(self, ctx: ModelParser.StructureContext):
        self.structure_name = ctx.name().getText()
        self.output_asp.append("")
        self.output_asp.append(f'structure("{self.structure_name}").')
        super().visitStructure(ctx)
        self.structure_name = self.root_name

    def visitEnumeration(self, ctx: ModelParser.EnumerationContext):
        self.parent_enum = ctx
        self.output_asp.append("")
        self.output_asp.append(f'enumeration("{ctx.name().getText()}").')
        super().visitEnumeration(ctx)
        self.parent_enum = None

    def visitBehavior(self, ctx: ModelParser.BehaviorContext):
        if ctx.name() is not None:
            self.context = ctx.name().getText()
        super().visitBehavior(ctx)
        self.context = self.root_name

    def visitFeature(self, ctx: ModelParser.FeatureContext):
        field: ModelParser.FieldContext = ctx.field()
        feature_name = field.fieldName.getText()

        if field.number_def() is not None:
            type_name = "num"
        elif field.string_def() is not None:
            return  # ignore string features
        elif field.type_ref is not None:
            type_name = field.type_ref.NAME()
        # else:
        #     type_name = feature_name

        cardinality: ModelParser.CardinalityContext = ctx.cardinality()
        c_min = 1
        c_max = 1
        if cardinality is not None:
            c_min = cardinality.min_.text.replace("x", "")
            c_max = c_min
            if cardinality.max_ is not None:
                c_max = cardinality.max_.text.replace("x", "").replace("*", "#sup")

        self.output_asp.append(f'feature("{self.structure_name}","{feature_name}","{type_name}",{c_min},{c_max}).')
        if type_name == "num":
            num: ModelParser.Number_defContext = field.number_def()
            if num.min_ is not None or num.max_ is not None:
                r_min = "#inf" if num.min_.getText() == "-\u221e" else num.min_.getText()  # negative infinity symbol
                r_max = "#sup" if num.max_.getText() == "\u221e" else num.max_.getText()  # infinity symbol
                self.output_asp.append(f'range("{self.structure_name}","{feature_name}",{r_min},{r_max}).')

    def visitAttribute(self, ctx: ModelParser.AttributeContext):
        # if self.parent_enum is None:
        #     raise ValueError("illegal option")
        parent_name = self.parent_enum.name().getText()
        field: ModelParser.FieldContext = ctx.field()
        if field.number_def() is not None:
            field_type = "num"
        else:
            field_type = "str"
        field_name = field.fieldName.getText()
        self.output_asp.append(f'attribute("{parent_name}","{field_name}","{field_type}").')
        super().visitAttribute(ctx)

    def visitOption(self, ctx: ModelParser.OptionContext):
        # if self.parent_enum is None:
        #     raise ValueError("illegal option")
        parent_name = self.parent_enum.name().getText()
        option_name = ctx.name().getText()
        self.output_asp.append(f'option("{parent_name}", "{option_name}").')

        constant: ModelParser.ConstantContext = ctx.constant()
        if constant != []:
            parent_attr: ModelParser.AttributeContext = self.parent_enum.attribute()
            for a, c in zip(parent_attr, constant):
                field: ModelParser.FieldContext = a.field()
                attr_name = field.fieldName.getText()
                if c.floating() is not None:
                    option_value = c.floating().getText()
                elif c.name() is not None:
                    option_value = f'"{c.name().getText()}"'
                self.output_asp.append(
                    f'attribute_value("{parent_name}","{option_name}","{attr_name}",{option_value}).'
                )

    def visitConditioned(self, ctx: ModelParser.ConditionedContext):
        self.condition_idx = 0
        if ctx.interaction() is None:
            self.output_asp.append("")
            self.output_asp.append(f"behavior({self.constraint_idx}).")
            self.output_asp.append(f'context({self.constraint_idx},"{self.context}").')
            super().visitConditioned(ctx)
            self.constraint_idx += 1

    def visitExplanation(self, ctx: ModelParser.ExplanationContext):
        self.output_asp.append(f"explanation({self.constraint_idx},{ctx.name().getText()}).")
        return super().visitExplanation(ctx)

    def visitAssign_default(self, ctx: ModelParser.Assign_defaultContext):
        path = ctx.path().getText()
        formula = ctx.formula().getText()
        self.output_asp.append(f'default({self.constraint_idx},"{path}","{formula}").')
        super().visitAssign_default(ctx)

    def visitAssign_imply(self, ctx: ModelParser.Assign_implyContext):
        path = ctx.path().getText()
        formula = ctx.formula().getText()
        self.output_asp.append(f'imply({self.constraint_idx},"{path}","{formula}").')
        super().visitAssign_imply(ctx)

    def visitCombinations(self, ctx: ModelParser.CombinationsContext):
        for i, f in enumerate(ctx.formula()):
            self.output_asp.append(f'combinations({self.constraint_idx},{i},"{f.getText()}").')
        super().visitCombinations(ctx)
        self.row_idx = 0

    def visitCombination_row(self, ctx: ModelParser.Combination_rowContext):
        row_type = ctx.rowType.text
        for col_idx, item in enumerate(ctx.combination_item()):
            values = item.getText()
            # Removing brackets around the values. Is this safe?
            if "," in values:
                values = values[1:-1]
            for v in values.split(","):
                if v == "-*-":  # Wildcard operator for combinations table
                    continue
                self.output_asp.append(f'{row_type}({self.constraint_idx},({col_idx},{self.row_idx}),"{v}").')
        self.print_path = False
        super().visitCombination_row(ctx)
        self.print_path = True
        self.row_idx += 1

    def visitPrecondition(self, ctx: ModelParser.PreconditionContext):
        condition = f'"{ctx.condition().getText()}"'
        self.output_asp.append(f"condition({self.constraint_idx},{self.condition_idx},{condition}).")
        self.condition_idx += 1
        super().visitPrecondition(ctx)

    def visitRequire(self, ctx: ModelParser.RequireContext):
        condition = f'"{ctx.condition().getText()}"'
        self.output_asp.append(f"require({self.constraint_idx},{condition}).")
        super().visitRequire(ctx)

    def visitCondition_or(self, ctx: ModelParser.Condition_orContext):
        cond_and: ModelParser.condition_andContext = ctx.condition_and()
        for i in range(len(cond_and) - 1):
            left = cond_and[i].getText()
            right = "||".join([a.getText() for a in cond_and[i + 1 :]])
            complete = left + "||" + right
            self.output_asp.append(f'binary("{complete}","{left}","||","{right}").')
        super().visitCondition_or(ctx)

    def visitCondition_and(self, ctx: ModelParser.Condition_andContext):
        cond_not: ModelParser.condition_notContext = ctx.condition_not()
        for i in range(len(cond_not) - 1):
            left = cond_not[i].getText()
            right = "&&".join([a.getText() for a in cond_not[i + 1 :]])
            complete = left + "&&" + right
            self.output_asp.append(f'binary("{complete}","{left}","&&","{right}").')
        super().visitCondition_and(ctx)

    def visitCondition_not(self, ctx: ModelParser.Condition_notContext):
        complete = ctx.getText()
        if ctx.condition_not() is not None:
            negated = ctx.condition_not().getText()
            self.output_asp.append(f'unary("{complete}","!","{negated}").')
        elif ctx.condition() is not None:
            in_brackets = ctx.condition().getText()
            self.output_asp.append(f'unary("{complete}","()","{in_brackets}").')
        super().visitCondition_not(ctx)

    def visitCondition_compare(self, ctx: ModelParser.Condition_compareContext):
        formula: ModelParser.FormulaContext = ctx.formula()
        parts: ModelParser.Condition_partContext = ctx.condition_part()

        left = formula.getText()
        for i, p in enumerate(parts):
            # Binary atom for compare
            right = p.formula().getText()
            compare = p.compare().getText()
            complete = left + compare + right
            self.output_asp.append(f'binary("{complete}","{left}","{compare}","{right}").')
            left = right

            # # For multiple comparisons rewrite as propositional formulas connected by &&
            # right_prop = "&&".join(
            #     [f"{l.formula().getText()}{r.getText()}" for l, r in (zip(parts[i:], parts[i + 1 :]))]
            # )
            # if right_prop != "":
            #     complete_prop = complete + "&&" + right_prop
            #     self.output_asp.append(f'binary("{complete_prop}","{complete}","&&","{right_prop}").')
        super().visitCondition_compare(ctx)

    def visitOptimize(self, ctx: ModelParser.OptimizeContext):
        priority = 0
        if ctx.priority:
            priority = int(ctx.priority.text)
        self.output_asp.append(f'{ctx.op.text}({self.constraint_idx},{priority},"{ctx.formula().getText()}").')
        return super().visitOptimize(ctx)

    def visitFormula_add(self, ctx: ModelParser.Formula_addContext):
        form_sub: ModelParser.Formula_subContext = ctx.formula_sub()
        for i in range(len(form_sub) - 1):
            left = form_sub[i].getText()
            right = "+".join([a.getText() for a in form_sub[i + 1 :]])
            complete = left + "+" + right
            self.output_asp.append(f'binary("{complete}","{left}","+","{right}").')
        super().visitFormula_add(ctx)

    def visitFormula_sub(self, ctx: ModelParser.Formula_subContext):
        form_mul: ModelParser.Formula_mulContext = ctx.formula_mul()
        for i in range(len(form_mul) - 1):
            left = form_mul[i].getText()
            right = "-".join([a.getText() for a in form_mul[i + 1 :]])
            complete = left + "-" + right
            self.output_asp.append(f'binary("{complete}","{left}","-","{right}").')
        super().visitFormula_sub(ctx)

    def visitFormula_mul(self, ctx: ModelParser.Formula_mulContext):
        form_div: ModelParser.Formula_divContext = ctx.formula_div()
        for i in range(len(form_div) - 1):
            left = form_div[i].getText()
            right = "*".join([a.getText() for a in form_div[i + 1 :]])
            complete = left + "*" + right
            self.output_asp.append(f'binary("{complete}","{left}","*","{right}").')
        super().visitFormula_mul(ctx)

    def visitFormula_div(self, ctx: ModelParser.Formula_divContext):
        form_pow: ModelParser.Formula_powContext = ctx.formula_pow()
        for i in range(len(form_pow) - 1):
            left = form_pow[i].getText()
            right = "/".join([a.getText() for a in form_pow[i + 1 :]])
            complete = left + "/" + right
            self.output_asp.append(f'binary("{complete}","{left}","/","{right}").')
        super().visitFormula_div(ctx)

    def visitFormula_pow(self, ctx: ModelParser.Formula_powContext):
        form_sign: ModelParser.Formula_signContext = ctx.formula_sign()
        for i in range(len(form_sign) - 1):
            left = form_sign[i].getText()
            right = "^".join([a.getText() for a in form_sign[i + 1 :]])
            complete = left + "^" + right
            self.output_asp.append(f'binary("{complete}","{left}","^","{right}").')
        super().visitFormula_pow(ctx)

    def visitFormula_sign(self, ctx: ModelParser.Formula_signContext):
        complete = ctx.getText()
        if ctx.formula_sign() is not None:
            if ctx.neg is not None:
                negated = ctx.formula_sign().getText()
                self.output_asp.append(f'unary("{complete}","-","{negated}").')
            else:
                # Is this really necessary?
                positive = ctx.formula_sign().getText()
                self.output_asp.append(f'unary("{complete}","+","{positive}").')
        elif ctx.formula() is not None:
            in_brackets = ctx.formula().getText()
            self.output_asp.append(f'unary("{complete}","()","{in_brackets}").')
        elif ctx.formula_func() is not None:
            func = ctx.formula_func().FUNCTION()
            for f in ctx.formula_func().formula():
                if str(func) in ["sum", "count", "min", "max", "avg"]:
                    self.output_asp.append(f'function("{self.context}","{complete}","{func}","{f.getText()}").')
                else:
                    self.output_asp.append(f'unary("{complete}","{func}","{f.getText()}").')
        super().visitFormula_sign(ctx)

    def visitPath(self, ctx: ModelParser.PathContext):
        # Only do this for actual paths? Not formulas
        if self.print_path:
            full_path = f"{ctx.getText()}"

            if full_path[0].isupper():
                self.output_asp.append(f'constant("{full_path}").')
            else:
                for i, p in enumerate(ctx.path_item()):
                    self.output_asp.append(f'path("{full_path}",{i},"{p.getText()}").')

    def visitFloating(self, ctx: ModelParser.FloatingContext):
        # if ctx.FLOATING() is not None:
        #     pass
        if ctx.INTEGER() is not None:
            self.output_asp.append(f'number("{ctx.INTEGER()}",{ctx.INTEGER()}).')



"""
Basic functions to run tests.
"""

import tempfile
from copy import deepcopy
from os.path import join
from typing import Any, Callable, List, Optional, Sequence, Tuple

from antlr4 import InputStream
from clingo import Application, Control, Symbol
from clintest.solver import Clingo, Solver
from clintest.test import Context, Record, Test

from coomsuite.application import COOMSolverApp
from coomsuite.preprocess import preprocess
from coomsuite.utils import run_antlr4_visitor


def parse_coom(coom_input: str, grammar: str = "model") -> List[str]:
    """
    Helper function for testing the COOM to ASP parser.
    """
    input_stream = InputStream(coom_input)
    asp_facts = run_antlr4_visitor(input_stream, grammar=grammar)
    return [a for a in asp_facts if a != ""]


def unpack_test(test_name: str, tests: dict[str, Any], flingo: bool = False) -> Tuple[Any, Any, Any]:
    """
    Unpacks a clintest.Test with parameters in a dictionary.

    Args:
        test_name (str): The dictionary key of the test.
    """
    test_dict = tests[test_name]
    program = test_dict.get("program", None)
    files = test_dict.get("files", [f"{test_name}.lp"]) if program is None else None
    if flingo:
        test = test_dict.get("ftest", test_dict["test"])
    else:
        test = test_dict["test"]
    test_with_name = Record(Context(test, str_=lambda test: f"{test_name} \n\n {str(test)}"))
    return test_with_name, program, files


def get_model_from_file(model_file: str) -> set[Symbol | str]:
    """
    Helper function to get the model for a clintest from a file.

    Args:
        model_file (str): The file containing the model in the directory tests/clintests/results
    """
    model: set[Symbol | str] = set()
    with open(join("tests", "clintests", "results", model_file), "r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if line:
                model.add(line[:-1])

    return model


def run_test(
    test: Test,
    files: Optional[List[str]] = None,
    program: Optional[str] = None,
    ctl_args: Optional[List[str]] = None,
    **kwargs: Any,
) -> None:
    """Creates a solver and runs a clintest test.

    Args:
        test (clintest.test.Test): The clintest test
        files (Optional[List[str]], optional): Test input filename saved in examples/tests
        program (Optional[str], optional): A clingo program. Defaults to ""
        ctl_args (Optional[List[str]], optional): List of arguments for clingo.Control. Defaults to [].
    """
    solver = kwargs.get("solver", "clingo")
    is_preprocess = solver == "preprocess"
    options = {
        "solver": solver,
        "output_format": kwargs.get("output_format", "asp"),
    }
    file_paths = (
        [join("examples", "tests", "solve" if not is_preprocess else "preprocess", f) for f in files] if files else []
    )
    if program:
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as tmp:
            tmp_name = tmp.name
            tmp.write(program)
        file_paths.append(tmp_name)
    ctl_args = [] if ctl_args is None else ctl_args

    if is_preprocess:
        multishot = kwargs.get("multishot", False)
        max_bound = kwargs.get("max_bound", 99)
        solver = Clingo(
            program="".join(preprocess(file_paths, discrete=False, max_bound=max_bound, multishot=multishot))
        )
    else:
        coom_app = COOMSolverApp(options=options, istest=True)
        solver = AppSolver(application=coom_app, files=file_paths, arguments=ctl_args)

    test_copy = deepcopy(test)
    solver.solve(test_copy)
    test_copy.assert_()


def compose(on_app: Callable, on_test: Callable) -> Callable:  # type: ignore
    """
    Composes two functions
    Args:
        on_app (Callable): Function from the application class
        on_test (Callable): Function for the test
    """

    def f(*args):  # type: ignore
        """Composed function"""
        if on_app is not None:
            on_app(*args)
        on_test(*args)

    return f


class MockControl:
    """
    Mocks a clingo control object to call the Application class
    """

    def __init__(
        self,
        test: Test,
        arguments: Optional[Sequence[str]] = None,
    ) -> None:
        arguments = [] if arguments is None else arguments
        self._test = test
        self._ctl = Control(arguments=arguments)

    def __getattr__(self, attr):  # type: ignore
        try:
            return self._ctl.__getattribute__(attr)
        except AttributeError:
            return self.__get_global_handler(attr)

    def solve(  # type: ignore # pylint: disable=too-many-arguments, too-many-positional-arguments
        self,
        on_model=None,
        on_unsat=None,
        on_core=None,
        on_statistics=None,
        on_finish=None,
    ) -> None:
        """
        Replaced the solve of the control by calling first internal callbacks and then
        the ones from the test
        """
        if not self._test.outcome().is_certain():
            self._ctl.solve(
                on_model=compose(on_model, self._test.on_model),
                on_unsat=compose(on_unsat, self._test.on_unsat),
                on_core=compose(on_core, self._test.on_core),
                on_statistics=compose(on_statistics, self._test.on_statistics),
                on_finish=compose(on_finish, self._test.on_finish),
            )


class AppSolver(Solver):
    """
    A clintest solver wrapping an application class
    """

    def __init__(
        self,
        application: Application,
        files: List[str],
        arguments: Optional[List[str]] = None,
    ) -> None:
        self.__application = application
        self.__arguments = [] if arguments is None else arguments
        self.__files = files

    def solve(self, test: Test) -> None:
        """Solves with clintest."""
        ctl = MockControl(test, self.__arguments)
        self.__application.main(control=ctl, files=self.__files)  # type: ignore



# Generated from /home/heuer/tools/coom-suite/src/coomsuite/utils/coom_grammar/user/UserInput.g4 by ANTLR 4.13.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,23,101,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,2,9,7,9,2,10,7,10,2,11,7,11,1,0,1,0,5,0,27,8,0,
        10,0,12,0,30,9,0,1,0,1,0,1,1,1,1,1,1,1,1,5,1,38,8,1,10,1,12,1,41,
        9,1,1,1,1,1,1,2,1,2,3,2,47,8,2,1,3,1,3,1,3,1,3,1,3,1,4,1,4,1,4,1,
        5,3,5,58,8,5,1,5,1,5,1,6,1,6,1,6,5,6,65,8,6,10,6,12,6,68,9,6,1,7,
        1,7,1,7,1,7,1,7,3,7,75,8,7,1,7,1,7,3,7,79,8,7,1,8,1,8,1,8,1,8,3,
        8,85,8,8,3,8,87,8,8,1,9,1,9,1,10,1,10,3,10,93,8,10,1,11,1,11,1,11,
        1,11,3,11,99,8,11,1,11,0,0,12,0,2,4,6,8,10,12,14,16,18,20,22,0,1,
        2,0,8,8,18,19,102,0,28,1,0,0,0,2,33,1,0,0,0,4,46,1,0,0,0,6,48,1,
        0,0,0,8,53,1,0,0,0,10,57,1,0,0,0,12,61,1,0,0,0,14,69,1,0,0,0,16,
        86,1,0,0,0,18,88,1,0,0,0,20,92,1,0,0,0,22,98,1,0,0,0,24,27,3,2,1,
        0,25,27,3,4,2,0,26,24,1,0,0,0,26,25,1,0,0,0,27,30,1,0,0,0,28,26,
        1,0,0,0,28,29,1,0,0,0,29,31,1,0,0,0,30,28,1,0,0,0,31,32,5,0,0,1,
        32,1,1,0,0,0,33,34,5,1,0,0,34,35,3,12,6,0,35,39,5,2,0,0,36,38,3,
        4,2,0,37,36,1,0,0,0,38,41,1,0,0,0,39,37,1,0,0,0,39,40,1,0,0,0,40,
        42,1,0,0,0,41,39,1,0,0,0,42,43,5,3,0,0,43,3,1,0,0,0,44,47,3,6,3,
        0,45,47,3,8,4,0,46,44,1,0,0,0,46,45,1,0,0,0,47,5,1,0,0,0,48,49,5,
        4,0,0,49,50,3,12,6,0,50,51,5,5,0,0,51,52,3,22,11,0,52,7,1,0,0,0,
        53,54,5,6,0,0,54,55,3,12,6,0,55,9,1,0,0,0,56,58,5,7,0,0,57,56,1,
        0,0,0,57,58,1,0,0,0,58,59,1,0,0,0,59,60,7,0,0,0,60,11,1,0,0,0,61,
        66,3,14,7,0,62,63,5,9,0,0,63,65,3,14,7,0,64,62,1,0,0,0,65,68,1,0,
        0,0,66,64,1,0,0,0,66,67,1,0,0,0,67,13,1,0,0,0,68,66,1,0,0,0,69,78,
        3,18,9,0,70,71,5,10,0,0,71,74,3,16,8,0,72,73,5,11,0,0,73,75,3,16,
        8,0,74,72,1,0,0,0,74,75,1,0,0,0,75,76,1,0,0,0,76,77,5,12,0,0,77,
        79,1,0,0,0,78,70,1,0,0,0,78,79,1,0,0,0,79,15,1,0,0,0,80,87,5,18,
        0,0,81,84,5,13,0,0,82,83,5,7,0,0,83,85,5,18,0,0,84,82,1,0,0,0,84,
        85,1,0,0,0,85,87,1,0,0,0,86,80,1,0,0,0,86,81,1,0,0,0,87,17,1,0,0,
        0,88,89,5,17,0,0,89,19,1,0,0,0,90,93,5,14,0,0,91,93,6,10,-1,0,92,
        90,1,0,0,0,92,91,1,0,0,0,93,21,1,0,0,0,94,99,5,15,0,0,95,99,5,16,
        0,0,96,99,3,10,5,0,97,99,3,12,6,0,98,94,1,0,0,0,98,95,1,0,0,0,98,
        96,1,0,0,0,98,97,1,0,0,0,99,23,1,0,0,0,12,26,28,39,46,57,66,74,78,
        84,86,92,98
    ]

class UserInputParser ( Parser ):

    grammarFileName = "UserInput.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'blockinput'", "'{'", "'}'", "'set'", 
                     "'='", "'add'", "'-'", "'\\u221E'", "'.'", "'['", "'..'", 
                     "']'", "'last'", "';'", "'true'", "'false'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "NAME", "INTEGER", "FLOATING", "NEWLINE", 
                      "WHITESPACE", "COMMENT", "MULTILINE_COMMENT" ]

    RULE_user_input = 0
    RULE_input_block = 1
    RULE_input_operation = 2
    RULE_set_value = 3
    RULE_add_instance = 4
    RULE_floating = 5
    RULE_path = 6
    RULE_path_item = 7
    RULE_path_index = 8
    RULE_name = 9
    RULE_stmt_end = 10
    RULE_formula_atom = 11

    ruleNames =  [ "user_input", "input_block", "input_operation", "set_value", 
                   "add_instance", "floating", "path", "path_item", "path_index", 
                   "name", "stmt_end", "formula_atom" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    T__10=11
    T__11=12
    T__12=13
    T__13=14
    T__14=15
    T__15=16
    NAME=17
    INTEGER=18
    FLOATING=19
    NEWLINE=20
    WHITESPACE=21
    COMMENT=22
    MULTILINE_COMMENT=23

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    def wasNewline(self):
        for index in reversed(range(self.getCurrentToken().tokenIndex)):
            # stop on default channel
            token = self.getTokenStream().get(index)
            if token.channel == 0:
                break

            # if the token is blank and contains newline, we found it
            if len(token.text) == 0:
                continue
            if token.text.startswith("\n") or token.text.startswith("\r"):
                return True

        return False



    class User_inputContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(UserInputParser.EOF, 0)

        def input_block(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(UserInputParser.Input_blockContext)
            else:
                return self.getTypedRuleContext(UserInputParser.Input_blockContext,i)


        def input_operation(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(UserInputParser.Input_operationContext)
            else:
                return self.getTypedRuleContext(UserInputParser.Input_operationContext,i)


        def getRuleIndex(self):
            return UserInputParser.RULE_user_input

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterUser_input" ):
                listener.enterUser_input(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitUser_input" ):
                listener.exitUser_input(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitUser_input" ):
                return visitor.visitUser_input(self)
            else:
                return visitor.visitChildren(self)




    def user_input(self):

        localctx = UserInputParser.User_inputContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_user_input)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 28
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 82) != 0):
                self.state = 26
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [1]:
                    self.state = 24
                    self.input_block()
                    pass
                elif token in [4, 6]:
                    self.state = 25
                    self.input_operation()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 30
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 31
            self.match(UserInputParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Input_blockContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def path(self):
            return self.getTypedRuleContext(UserInputParser.PathContext,0)


        def input_operation(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(UserInputParser.Input_operationContext)
            else:
                return self.getTypedRuleContext(UserInputParser.Input_operationContext,i)


        def getRuleIndex(self):
            return UserInputParser.RULE_input_block

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterInput_block" ):
                listener.enterInput_block(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitInput_block" ):
                listener.exitInput_block(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitInput_block" ):
                return visitor.visitInput_block(self)
            else:
                return visitor.visitChildren(self)




    def input_block(self):

        localctx = UserInputParser.Input_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_input_block)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 33
            self.match(UserInputParser.T__0)
            self.state = 34
            self.path()
            self.state = 35
            self.match(UserInputParser.T__1)
            self.state = 39
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==4 or _la==6:
                self.state = 36
                self.input_operation()
                self.state = 41
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 42
            self.match(UserInputParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Input_operationContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def set_value(self):
            return self.getTypedRuleContext(UserInputParser.Set_valueContext,0)


        def add_instance(self):
            return self.getTypedRuleContext(UserInputParser.Add_instanceContext,0)


        def getRuleIndex(self):
            return UserInputParser.RULE_input_operation

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterInput_operation" ):
                listener.enterInput_operation(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitInput_operation" ):
                listener.exitInput_operation(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitInput_operation" ):
                return visitor.visitInput_operation(self)
            else:
                return visitor.visitChildren(self)




    def input_operation(self):

        localctx = UserInputParser.Input_operationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_input_operation)
        try:
            self.state = 46
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [4]:
                self.enterOuterAlt(localctx, 1)
                self.state = 44
                self.set_value()
                pass
            elif token in [6]:
                self.enterOuterAlt(localctx, 2)
                self.state = 45
                self.add_instance()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Set_valueContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.op = None # Token

        def path(self):
            return self.getTypedRuleContext(UserInputParser.PathContext,0)


        def formula_atom(self):
            return self.getTypedRuleContext(UserInputParser.Formula_atomContext,0)


        def getRuleIndex(self):
            return UserInputParser.RULE_set_value

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSet_value" ):
                listener.enterSet_value(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSet_value" ):
                listener.exitSet_value(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSet_value" ):
                return visitor.visitSet_value(self)
            else:
                return visitor.visitChildren(self)




    def set_value(self):

        localctx = UserInputParser.Set_valueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_set_value)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 48
            localctx.op = self.match(UserInputParser.T__3)
            self.state = 49
            self.path()
            self.state = 50
            self.match(UserInputParser.T__4)
            self.state = 51
            self.formula_atom()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Add_instanceContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.op = None # Token

        def path(self):
            return self.getTypedRuleContext(UserInputParser.PathContext,0)


        def getRuleIndex(self):
            return UserInputParser.RULE_add_instance

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAdd_instance" ):
                listener.enterAdd_instance(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAdd_instance" ):
                listener.exitAdd_instance(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAdd_instance" ):
                return visitor.visitAdd_instance(self)
            else:
                return visitor.visitChildren(self)




    def add_instance(self):

        localctx = UserInputParser.Add_instanceContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_add_instance)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 53
            localctx.op = self.match(UserInputParser.T__5)
            self.state = 54
            self.path()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class FloatingContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def FLOATING(self):
            return self.getToken(UserInputParser.FLOATING, 0)

        def INTEGER(self):
            return self.getToken(UserInputParser.INTEGER, 0)

        def getRuleIndex(self):
            return UserInputParser.RULE_floating

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFloating" ):
                listener.enterFloating(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFloating" ):
                listener.exitFloating(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitFloating" ):
                return visitor.visitFloating(self)
            else:
                return visitor.visitChildren(self)




    def floating(self):

        localctx = UserInputParser.FloatingContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_floating)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 57
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==7:
                self.state = 56
                self.match(UserInputParser.T__6)


            self.state = 59
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 786688) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PathContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def path_item(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(UserInputParser.Path_itemContext)
            else:
                return self.getTypedRuleContext(UserInputParser.Path_itemContext,i)


        def getRuleIndex(self):
            return UserInputParser.RULE_path

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPath" ):
                listener.enterPath(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPath" ):
                listener.exitPath(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPath" ):
                return visitor.visitPath(self)
            else:
                return visitor.visitChildren(self)




    def path(self):

        localctx = UserInputParser.PathContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_path)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 61
            self.path_item()
            self.state = 66
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==9:
                self.state = 62
                self.match(UserInputParser.T__8)
                self.state = 63
                self.path_item()
                self.state = 68
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Path_itemContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def name(self):
            return self.getTypedRuleContext(UserInputParser.NameContext,0)


        def path_index(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(UserInputParser.Path_indexContext)
            else:
                return self.getTypedRuleContext(UserInputParser.Path_indexContext,i)


        def getRuleIndex(self):
            return UserInputParser.RULE_path_item

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPath_item" ):
                listener.enterPath_item(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPath_item" ):
                listener.exitPath_item(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPath_item" ):
                return visitor.visitPath_item(self)
            else:
                return visitor.visitChildren(self)




    def path_item(self):

        localctx = UserInputParser.Path_itemContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_path_item)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 69
            self.name()
            self.state = 78
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==10:
                self.state = 70
                self.match(UserInputParser.T__9)
                self.state = 71
                self.path_index()
                self.state = 74
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==11:
                    self.state = 72
                    self.match(UserInputParser.T__10)
                    self.state = 73
                    self.path_index()


                self.state = 76
                self.match(UserInputParser.T__11)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Path_indexContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INTEGER(self):
            return self.getToken(UserInputParser.INTEGER, 0)

        def getRuleIndex(self):
            return UserInputParser.RULE_path_index

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPath_index" ):
                listener.enterPath_index(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPath_index" ):
                listener.exitPath_index(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPath_index" ):
                return visitor.visitPath_index(self)
            else:
                return visitor.visitChildren(self)




    def path_index(self):

        localctx = UserInputParser.Path_indexContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_path_index)
        self._la = 0 # Token type
        try:
            self.state = 86
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [18]:
                self.enterOuterAlt(localctx, 1)
                self.state = 80
                self.match(UserInputParser.INTEGER)
                pass
            elif token in [13]:
                self.enterOuterAlt(localctx, 2)
                self.state = 81
                self.match(UserInputParser.T__12)
                self.state = 84
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==7:
                    self.state = 82
                    self.match(UserInputParser.T__6)
                    self.state = 83
                    self.match(UserInputParser.INTEGER)


                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class NameContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(UserInputParser.NAME, 0)

        def getRuleIndex(self):
            return UserInputParser.RULE_name

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterName" ):
                listener.enterName(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitName" ):
                listener.exitName(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitName" ):
                return visitor.visitName(self)
            else:
                return visitor.visitChildren(self)




    def name(self):

        localctx = UserInputParser.NameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_name)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 88
            self.match(UserInputParser.NAME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Stmt_endContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return UserInputParser.RULE_stmt_end

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStmt_end" ):
                listener.enterStmt_end(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStmt_end" ):
                listener.exitStmt_end(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitStmt_end" ):
                return visitor.visitStmt_end(self)
            else:
                return visitor.visitChildren(self)




    def stmt_end(self):

        localctx = UserInputParser.Stmt_endContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_stmt_end)
        try:
            self.state = 92
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [14]:
                self.enterOuterAlt(localctx, 1)
                self.state = 90
                self.match(UserInputParser.T__13)
                pass
            elif token in [-1]:
                self.enterOuterAlt(localctx, 2)
                self.wasNewline()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Formula_atomContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.atom_true = None # Token
            self.atom_false = None # Token
            self.atom_num = None # FloatingContext
            self.atom_path = None # PathContext

        def floating(self):
            return self.getTypedRuleContext(UserInputParser.FloatingContext,0)


        def path(self):
            return self.getTypedRuleContext(UserInputParser.PathContext,0)


        def getRuleIndex(self):
            return UserInputParser.RULE_formula_atom

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFormula_atom" ):
                listener.enterFormula_atom(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFormula_atom" ):
                listener.exitFormula_atom(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitFormula_atom" ):
                return visitor.visitFormula_atom(self)
            else:
                return visitor.visitChildren(self)




    def formula_atom(self):

        localctx = UserInputParser.Formula_atomContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_formula_atom)
        try:
            self.state = 98
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [15]:
                self.enterOuterAlt(localctx, 1)
                self.state = 94
                localctx.atom_true = self.match(UserInputParser.T__14)
                pass
            elif token in [16]:
                self.enterOuterAlt(localctx, 2)
                self.state = 95
                localctx.atom_false = self.match(UserInputParser.T__15)
                pass
            elif token in [7, 8, 18, 19]:
                self.enterOuterAlt(localctx, 3)
                self.state = 96
                localctx.atom_num = self.floating()
                pass
            elif token in [17]:
                self.enterOuterAlt(localctx, 4)
                self.state = 97
                localctx.atom_path = self.path()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



"""
Test cases for preprocessing the serialized fact format.
"""

from typing import Any
from unittest import TestCase

from . import run_test, unpack_test
from .clintests.tests_preprocess import TESTS_PREPROCESS


class TestPreprocess(TestCase):
    """
    Test cases for the preprocessing encoding.
    """

    def run_test(self, test_name: str, **kwargs: Any) -> None:
        """
        Runs a clintest test with the preprocessing encoding (using clingo).
        """
        test, program, files = unpack_test(test_name, TESTS_PREPROCESS)
        run_test(test, files=files, program=program, ctl_args=["0"], solver="preprocess", **kwargs)

    def test_structure(self) -> None:
        """
        Test preprocessing COOM structures.
        """
        self.run_test("empty")
        self.run_test("empty_product")
        self.run_test("structure_mandatory")
        self.run_test("structure_optional")
        self.run_test("structure_nested")
        self.run_test("structure_nested_optional")

    def test_enumeration(self) -> None:
        """
        Test preprocessing COOM enumerations.
        """
        self.run_test("enumeration")
        self.run_test("bool_enumeration")
        self.run_test("attribute")

    def test_require(self) -> None:
        """
        Test preprocessing COOM require constraints.
        """
        self.run_test("require_undef")
        self.run_test("require_with_number")
        self.run_test("require_with_number_ge")
        self.run_test("require_with_constant")
        self.run_test("require_two_wheels")
        self.run_test("conditional_require_undef")
        self.run_test("conditional_require")
        self.run_test("multiple_conditions")

        self.run_test("require_multiple_instances")
        self.run_test("require_with_optional_part")
        self.run_test("require_with_partonomy")
        self.run_test("require_with_partonomy2")
        self.run_test("require_with_partonomy_multiple_instances")

    def test_combinations_table(self) -> None:
        """
        Test preprocessing COOM combinations tables.
        """
        self.run_test("combination")
        self.run_test("combination_with_structure")
        self.run_test("combination_at_part_with_wildcard")
        self.run_test("combination_at_part_multiple_instances")

    def test_numerics(self) -> None:
        """
        Test preprocessing numeric features, arithmetics and numerical function in COOM.
        """
        self.run_test("simple_numeric_feature")
        self.run_test("simple_arithmetic_plus")
        self.run_test("simple_arithmetic_minus")
        self.run_test("simple_arithmetic_multiplication")
        self.run_test("simple_arithmetic_plus_default_right")
        self.run_test("simple_arithmetic_plus_default_left")
        self.run_test("simple_arithmetic_minus_default_right")
        self.run_test("simple_arithmetic_minus_default_left")
        self.run_test("parentheses")

        self.run_test("count")
        self.run_test("sum")

    def test_defaults(self) -> None:
        """
        Test defaults in COOM
        """
        self.run_test("simple_default")

    def test_optimization(self) -> None:
        """
        Test optimization in COOM
        """

        self.run_test("minimize")
        self.run_test("maximize")
        self.run_test("minimize_non_root_path")
        self.run_test("maximize_non_root_path")
        self.run_test("minimize_function")
        self.run_test("maximize_function")

    def test_user_input(self) -> None:
        """
        Test preprocessing COOM user input.
        """
        self.run_test("set_constant")
        self.run_test("set_number")
        self.run_test("add")

    def test_unbounded_cardinalities(self) -> None:
        """
        Test preprocessing of unbounded cardinalities.
        """
        self.run_test("unbounded_singleshot_zero_lb_zero_max", max_bound=0)
        self.run_test("unbounded_singleshot_zero_lb_one_max", max_bound=1)
        self.run_test("unbounded_singleshot_two_lb_zero_max", max_bound=0)
        self.run_test("unbounded_singleshot_two_lb_one_max", max_bound=1)
        self.run_test("unbounded_singleshot_optimize", max_bound=0)

    def test_unbounded_cardinalities_multishot(self) -> None:
        """
        Test multishot-specific preprocessing of unbounded cardinalities.
        """
        # test that multishot preprocessing results in a superset of singleshot preprocessing
        self.run_test("unbounded_multishot_zero_lb_zero_max_superset", max_bound=0, multishot=True)
        self.run_test("unbounded_multishot_zero_lb_one_max_superset", max_bound=1, multishot=True)
        self.run_test("unbounded_multishot_two_lb_zero_max_superset", max_bound=0, multishot=True)
        self.run_test("unbounded_multishot_two_lb_one_max_superset", max_bound=1, multishot=True)
        self.run_test("unbounded_multishot_optimize_superset", max_bound=0, multishot=True)

        # test models of multishot preprocessing
        self.run_test("unbounded_multishot_zero_lb_zero_max", max_bound=0, multishot=True)
        self.run_test("unbounded_multishot_zero_lb_one_max", max_bound=1, multishot=True)
        self.run_test("unbounded_multishot_two_lb_zero_max", max_bound=0, multishot=True)
        self.run_test("unbounded_multishot_two_lb_one_max", max_bound=1, multishot=True)
        self.run_test("unbounded_multishot_optimize", max_bound=0, multishot=True)



"""
Contains a module for solving problems with open bounds/cardinalities.
"""

from typing import List, Optional

from clingo.application import clingo_main

from coomsuite import solve

from . import get_bound_iter, next_bound_converge
from .multi_application import COOMMultiSolverApp

# pylint: disable=line-too-long
# Reference for exit code: https://github.com/potassco/clasp/issues/42#issuecomment-459981038
ret_dict = {0: "UNKNOWN", 1: "INTERRUPT", 10: "SAT", 20: "UNSAT", 30: "OPT", 33: "MEMORY", 65: "ERROR", 128: "NO_RUN"}


class BoundSolver:
    """
    Module for solving problems with open bounds/cardinalities.
    """

    facts: List[str]
    solver: str
    clingo_args: List[str]
    output_format: str

    def __init__(
        self,
        facts: List[str],
        solver: str,
        clingo_args: List[str],
        output_format: str,
    ):
        self.facts = facts
        self.solver = solver
        self.clingo_args = clingo_args
        self.output_format = output_format

    def _solve(self, max_bound: int) -> int:  # nocoverage
        return solve(self.facts, self.solver, max_bound, self.clingo_args, self.output_format)

    def _converge(self, unsat_bound: int, sat_bound: int) -> Optional[int]:
        """
        Converge to the minimal bound given lower bound (unsat_bound) and upper bound (sat_bound)
        """
        while True:
            current_bound = next_bound_converge(unsat_bound, sat_bound)

            if current_bound is None:
                print("\nOptimal bound found")
                return sat_bound

            print("\nOptimal bound not yet found")
            print(f"Solving with bound = {current_bound}\n")

            ret = self._solve(current_bound)
            try:
                if ret_dict[ret] in ["SAT", "OPT"]:
                    sat_bound = current_bound
                elif ret_dict[ret] == "UNSAT":
                    unsat_bound = current_bound
                else:
                    print(f"\n Some error occured during solving with max bound = {current_bound}\n Exit code: {ret}\n")
                    return None
            except KeyError as exc:
                raise KeyError("Unknown exit code.") from exc

    def get_bounds(
        self, algorithm: str = "linear", initial_bound: int = 0, step: Optional[int] = None, use_multishot: bool = False
    ) -> Optional[int]:
        """
        Compute the minimal bound for the problem.
        """
        if step is None:
            step = 1 if algorithm == "linear" else 2

        # multi shot solving
        if use_multishot:  # nocoverage
            multishot_solver = COOMMultiSolverApp(
                serialized_facts=self.facts,
                step=step,
                initial_bound=initial_bound,
                algorithm=algorithm,
                options={
                    "solver": self.solver,
                    "output_format": self.output_format,
                },
            )

            clingo_main(
                multishot_solver,
                self.clingo_args,
            )

            return multishot_solver.current_max_bound

        # single shot solving
        bounds_iter = get_bound_iter(algorithm, initial_bound, step)
        current_max_bound = next(bounds_iter)
        prev_bound = -1
        while True:
            print(f"\nSolving with max_bound = {current_max_bound}\n")
            ret = self._solve(current_max_bound)
            try:
                if ret_dict[ret] in ["SAT", "OPT"]:
                    return self._converge(prev_bound, current_max_bound)
                if ret_dict[ret] == "UNSAT":
                    pass
                else:
                    print(
                        f"\n Some error occured during solving with max bound = {current_max_bound}\n Exit code: {ret}\n"
                    )
                    return None
            except KeyError as exc:
                raise KeyError("Unknown exit code.") from exc
            prev_bound = current_max_bound
            current_max_bound = next(bounds_iter)
